# Chambolle, A.
# An Algorithm for Total Variation Minimization and Applications.
# Journal of Mathematical Imaging and Vision 20, 89–97 (2004).
# https://doi.org/10.1023/B:JMIV.0000011325.36760.1e
#
# Implementation Notes:
# - λ is called λ and is not a scalar but a scalar field
# - pointwise feasibility constraint |p|<=λ instead of |p|<=1
# - B is introduced, the original Chambolle algorithm has B=Id

using Base: @_inline_meta
using StaticKernels
using StaticArrays: SVector
using LinearAlgebra

struct ChambolleAlgorithm{P} <: Algorithm{P}
    problem::P

    "fixed point inertia parameter"
    τ::Float64
    "anisotropic projection"
    anisotropic::Bool

    function ChambolleAlgorithm(problem; τ=inv(8 * normB(problem)), anisotropic=false)
        return new{typeof(problem)}(problem, τ, anisotropic)
    end
end

struct ChambolleState{A,T,R,S,Sv,K1,K2} <: State
    algorithm::A

    "matrix view on .f"
    p::T

    "matrix view on rv"
    r::R
    "matrix view on sv"
    s::S

    "dual variable as vector"
    rv::Sv
    "scalar temporary variable"
    sv::Sv

    "div(p) + g kernel"
    k1::K1
    "(p + (τ/normB)*grad(q))/(1 + (τ/normB)/λ|grad(q)|) kernel"
    k2::K2
end

projnorm(v) = projnorm(v, false)
projnorm(v::SVector, anisotropic) = anisotropic ? abs.(v) : norm(vec(v))
#const SVectorNest = SArray{<:Any,<:SArray{<:Any,Float64,1},1}
# TODO: respect anisotropic
@inline projnorm(v, _) = norm(v)
#@inline projnorm(v::SVectorNest, _) = sqrt(sum(sum(v[i] .^ 2) for i in eachindex(v)))


function init(alg::ChambolleAlgorithm{<:DualTVL1ROFOpProblem})
    g = alg.problem.g
    λ = alg.problem.λ
    d = ndims(g)
    ax = axes(g)

    # TODO: generalize to SArray
    p1type(T::Type{<:Real}) = SVector{d,T}
    p1type(::Type{SVector{m,T}}) where {m, T} = SMatrix{m,d,T,m*d}

    rv = zeros(eltype(g), length(g))
    sv = zero(rv)

    p = extend(zeros(p1type(eltype(g)), ax), StaticKernels.ExtensionNothing())
    r = reshape(rv, size(g))
    s = extend(reshape(sv, size(g)), StaticKernels.ExtensionReplicate())

    @inline kf1(pw) = @inbounds divergence(pw) + g[pw.position]
    k1 = Kernel{ntuple(_->-1:1, d)}(kf1)

    @inline function kf2(sw)
        @inbounds iszero(λ[sw.position]) && return zero(p[sw.position])
        sgrad = alg.τ * gradient(sw)
        return @inbounds (p[sw.position] + sgrad) ./ (1 .+ projnorm(sgrad, alg.anisotropic) ./ λ[sw.position])
    end
    k2 = Kernel{ntuple(_->0:1, d)}(kf2)

    return ChambolleState(alg, p, r, s, rv, sv, k1, k2)
end

function reset!(ctx::ChambolleState)
    fill!(ctx.p, zero(eltype(ctx.p)))
end

# special case for diagonal ops
function LinearAlgebra.mul!(Y::AbstractVector{<:SVector}, A, B::AbstractVector{<:SVector})
    for i in eachindex(Y)
        Y[i] = A[i,i] * B[i]
    end
end

function step!(ctx::ChambolleState)
    alg = ctx.algorithm
    d = ndims(ctx.r)
    kgrad = Kernel{ntuple(_->0:1, d)}(gradient)

    λ = alg.problem.λ
    p = ctx.p

    sv = vec(parent(ctx.s))
    rv = vec(ctx.r)

    # r = div(p) + g
    map!(ctx.k1, ctx.r, ctx.p)
    # s = B * r
    mul!(sv, alg.problem.B, rv)
    # p = (p + τ*grad(s)) / (1 + τ/λ|grad(s)|)

    sgrad = map(kgrad, ctx.s)
    f(p, sgrad, λ) = iszero(λ) ? zero(p) :
        (p + alg.τ * sgrad) ./ (1. .+ projnorm(alg.τ * sgrad, alg.anisotropic) ./ λ)

    ctx.p .= f.(ctx.p, sgrad, λ)

    return ctx
end

fetch(ctx::ChambolleState) = ctx.p
