using LinearAlgebra: I

struct SurrogateAlgorithm{P} <: Algorithm{P}
    problem::P
    subalg::Function
    "number of inner iterations"
    ninner::Int
    τ::Float64
    function SurrogateAlgorithm(problem::DualTVL1ROFOpProblem;
                                ninner=1,
                                τ=2*normB(problem), subalg=x->ProjGradAlgorithm(x))
        return new{typeof(problem)}(problem, subalg, ninner, τ)
    end
end

struct SurrogateState{A,SS,Wv,R,S,K1} <: State
    algorithm::A
    substate::SS

    "scalar temporary 1"
    rv::Wv
    "scalar temporary 2"
    sv::Wv

    "matrix view on rv"
    r::R
    "matrix view on sv"
    s::S

    k1::K1
end

function init(alg::SurrogateAlgorithm)
    prob = alg.problem
    g = prob.g
    d = ndims(g)

    rv = zeros(eltype(g), length(g))
    sv = zero(rv)

    r = reshape(rv, size(g))
    s = reshape(sv, size(g))

    @inline kf1(pw) = @inbounds -divergence(pw)
    k1 = Kernel{ntuple(_->-1:1, d)}(kf1)

    subg = copy(g)
    subprob = DualTVL1ROFOpProblem(subg, I, prob.λ)
    subalg = alg.subalg(subprob)
    substate = init(subalg)

    # we don't want to modify ourselves if alg.subalg gives us a bad reference
    @assert subalg.problem.g !== g

    return SurrogateState(alg, substate, rv, sv, r, s, k1)
end

function update_g!(subg, st::SurrogateState)
    alg = st.algorithm
    g = alg.problem.g

    p = extend(fetch(st.substate), StaticKernels.ExtensionNothing())

    sv = vec(parent(st.s))
    rv = vec(st.r)

    # r = Λ*p
    map!(st.k1, st.r, p)
    # s = r - g
    st.s .= st.r .- g
    # r = r - 1/τ * B * s
    mul!(rv, alg.problem.B, sv, -1. / alg.τ, 1.)

    subg .= st.r
end


function step!(st::SurrogateState)
    subst = st.substate

    # update data g of the subproblem
    update_g!(subst.algorithm.problem.g, st)

    for i in 1:st.algorithm.ninner
        step!(subst)
    end

    return st
end

fetch(st::SurrogateState) = fetch(st.substate)
